/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_dpa.h>
#include <doca_log.h>
#include <doca_types.h>
#include <doca_argp.h>
#include <doca_sync_event.h>
#include <doca_mmap.h>
#include <doca_rdma.h>
#include <doca_ctx.h>

#include "dpa_ping_pong_core.h"
#include "../common/dpa_ping_pong_common_defs.h"

DOCA_LOG_REGISTER(DPA_PING_PONG::CORE);

/*
 * A struct that includes all needed info on registered kernels and is initialized during linkage by DPACC.
 * Variable name should be the token passed to DPACC with --app-name parameter.
 */
extern struct doca_dpa_app *dpa_ping_pong_app;

/**
 * kernels/RPC declaration
 */
doca_dpa_func_t thread_kernel;
doca_dpa_func_t trigger_first_iteration_rpc;

/**
 * @brief Mask for doca_sync_event_wait_gt() wait value
 */
#define SYNC_EVENT_MASK_FFS (0xFFFFFFFFFFFFFFFF)

/**
 * @brief A struct that includes all the resources needed for DPA thread
 */
struct dpa_thread_obj {
	struct doca_dpa *doca_dpa;	    /**< DOCA DPA context */
	doca_dpa_func_t *func;		    /**< DPA thread entry point */
	uint64_t arg;			    /**< DPA thread entry point argument */
	doca_dpa_dev_uintptr_t tls_dev_ptr; /**< DPA thread local storage device memory pointer */
	struct doca_dpa_thread *thread;	    /**< Created DPA thread */
};

/**
 * @brief A struct that includes all the resources needed for DPA completion
 */
struct dpa_completion_obj {
	struct doca_dpa *doca_dpa;	      /**< DOCA DPA context */
	unsigned int queue_size;	      /**< DPA completion queue size */
	struct doca_dpa_thread *thread;	      /**< DPA completion attached thread */
	struct doca_dpa_completion *dpa_comp; /**< Created DPA completion */
	doca_dpa_dev_completion_t handle;     /**< Created DPA completion device handle */
};

/**
 * @brief A struct that includes all the resources needed for DPA RDMA
 */
struct dpa_rdma_obj {
	struct doca_dev *doca_device;	/**< DOCA device for DPA */
	struct doca_dpa *doca_dpa;	/**< DOCA DPA context */
	uint32_t permissions;		/**< RDMA permissions */
	union doca_data user_data;	/**< RDMA user data */
	struct doca_rdma *rdma;		/**< Created RDMA */
	struct doca_ctx *rdma_as_ctx;	/**< Created RDMA context */
	doca_dpa_dev_rdma_t dpa_rdma;	/**< Created RDMA DPA device handle */
	const void *connection_details; /**< Created RDMA connection details from export */
	size_t conn_det_len;		/**< Created RDMA connection details length from export */
};

/**
 * @brief A struct that includes all the resources needed for DOCA Mmap
 */
struct doca_mmap_obj {
	struct doca_dev *doca_device;	     /**< DOCA device for DPA */
	struct doca_dpa *doca_dpa;	     /**< DOCA DPA context */
	uint32_t permissions;		     /**< Mmap permissions */
	void *memrange_addr;		     /**< Mmap address */
	size_t memrange_len;		     /**< Mmap address length */
	struct doca_mmap *mmap;		     /**< Created Mmap */
	doca_dpa_dev_mmap_t dpa_mmap_handle; /**< Created Mmap DPA device handle */
	const void *rdma_export;	     /**< Exported Mmap to be used for RDMA operations */
	size_t export_len;		     /**< Exported Mmap length */
};

/**
 * Check if the provided device name is a name of a valid device, exists and supports DPA
 *
 * @device_name [in]: The provided device name
 * @return: True if device_name is a valid device, false otherwise.
 */
static bool dpa_device_exists_check(const char *device_name)
{
	struct doca_devinfo **dev_list;
	uint32_t nb_devs = 0;
	doca_error_t result;
	bool exists = false;
	char ibdev_name[DOCA_DEVINFO_IBDEV_NAME_SIZE] = {0};

	if (strncmp(device_name, DEVICE_DEFAULT_NAME, strlen(DEVICE_DEFAULT_NAME)) == 0) {
		return true;
	}

	result = doca_devinfo_create_list(&dev_list, &nb_devs);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to load DOCA devices list: %s", doca_error_get_descr(result));
		return false;
	}

	for (uint32_t i = 0; i < nb_devs; i++) {
		result = doca_dpa_cap_is_supported(dev_list[i]);
		if (result != DOCA_SUCCESS) {
			continue;
		}

		result = doca_devinfo_get_ibdev_name(dev_list[i], ibdev_name, sizeof(ibdev_name));
		if (result != DOCA_SUCCESS) {
			continue;
		}

		if (strncmp(device_name, ibdev_name, DOCA_DEVINFO_IBDEV_NAME_SIZE) == 0) {
			exists = true;
			break;
		}
	}

	doca_devinfo_destroy_list(dev_list);

	return exists;
}

/**
 * ARGP Callback - Handle device name parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t device_name_param_callback(void *param, void *config)
{
	struct dpa_config *dpa_cgf = (struct dpa_config *)config;
	char *device_name = (char *)param;
	int len;

	len = strnlen(device_name, DOCA_DEVINFO_IBDEV_NAME_SIZE);
	if (len == DOCA_DEVINFO_IBDEV_NAME_SIZE) {
		DOCA_LOG_ERR("Entered device name exceeding the maximum size of %d", DOCA_DEVINFO_IBDEV_NAME_SIZE - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strncpy(dpa_cgf->device_name, device_name, len + 1);

	if (!dpa_device_exists_check(dpa_cgf->device_name)) {
		DOCA_LOG_ERR("Entered device name: %s doesn't exist or doesn't support DPA", dpa_cgf->device_name);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

/**
 * Register DPA application parameters
 *
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t register_dpa_params(void)
{
	doca_error_t result;
	struct doca_argp_param *device_param;

	result = doca_argp_param_create(&device_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_error_get_descr(result));
		return result;
	}
	doca_argp_param_set_short_name(device_param, "d");
	doca_argp_param_set_long_name(device_param, "device");
	doca_argp_param_set_arguments(device_param, "<device name>");
	doca_argp_param_set_description(
		device_param,
		"device name that supports DPA (optional). If not provided then a random device will be chosen");
	doca_argp_param_set_callback(device_param, device_name_param_callback);
	doca_argp_param_set_type(device_param, DOCA_ARGP_TYPE_STRING);
	result = doca_argp_register_param(device_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_error_get_descr(result));
		return result;
	}

	return DOCA_SUCCESS;
}

/**
 * Open DPA DOCA device
 *
 * @device_name [in]: Wanted device name, can be NOT_SET and then a random DPA supported device is chosen
 * @doca_device [out]: An allocated DOCA DPA device on success and NULL otherwise
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t open_dpa_device(const char *device_name, struct doca_dev **doca_device)
{
	struct doca_devinfo **dev_list;
	uint32_t nb_devs = 0;
	doca_error_t result;
	char ibdev_name[DOCA_DEVINFO_IBDEV_NAME_SIZE] = {0};
	uint32_t i = 0;

	result = doca_devinfo_create_list(&dev_list, &nb_devs);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to load DOCA devices list: %s", doca_error_get_descr(result));
		return result;
	}

	for (i = 0; i < nb_devs; i++) {
		result = doca_dpa_cap_is_supported(dev_list[i]);
		if (result != DOCA_SUCCESS) {
			continue;
		}

		result = doca_devinfo_get_ibdev_name(dev_list[i], ibdev_name, sizeof(ibdev_name));
		if (result != DOCA_SUCCESS ||
		    (strcmp(device_name, DEVICE_DEFAULT_NAME) != 0 && strcmp(device_name, ibdev_name) != 0)) {
			continue;
		}

		result = doca_dev_open(dev_list[i], doca_device);
		if (result != DOCA_SUCCESS) {
			doca_devinfo_destroy_list(dev_list);
			DOCA_LOG_ERR("Failed to open DOCA device: %s", doca_error_get_descr(result));
			return result;
		}

		break;
	}

	doca_devinfo_destroy_list(dev_list);

	if (*doca_device == NULL) {
		DOCA_LOG_ERR("Couldn't get DOCA device");
		return DOCA_ERROR_NOT_FOUND;
	}

	return result;
}

/**
 * @brief Allocate DPA resources, DPA context and DOCA device
 *
 * @cfg [in]: DPA config object
 * @resources [out]: DPA resources object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t allocate_dpa_resources(struct dpa_config *cfg, struct dpa_resources *resources)
{
	doca_error_t result;

	result = open_dpa_device(cfg->device_name, &resources->doca_device);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function open_dpa_device() failed");
		goto exit_label;
	}

	result = doca_dpa_create(resources->doca_device, &(resources->doca_dpa));
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA DPA context: %s", doca_error_get_descr(result));
		goto close_doca_dev;
	}

	result = doca_dpa_set_app(resources->doca_dpa, dpa_ping_pong_app);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set DOCA DPA app: %s", doca_error_get_descr(result));
		goto destroy_doca_dpa;
	}

	result = doca_dpa_start(resources->doca_dpa);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to start DOCA DPA context: %s", doca_error_get_descr(result));
		goto destroy_doca_dpa;
	}

	return result;

destroy_doca_dpa:
	doca_dpa_destroy(resources->doca_dpa);
close_doca_dev:
	doca_dev_close(resources->doca_device);
exit_label:
	return result;
}

/**
 * @brief Destroy DPA resources, DPA context and DOCA device
 *
 * @resources [in]: DPA resources object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t destroy_dpa_resources(struct dpa_resources *resources)
{
	doca_error_t result = DOCA_SUCCESS;
	doca_error_t tmp_result;

	tmp_result = doca_dpa_destroy(resources->doca_dpa);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_destroy() failed: %s", doca_error_get_descr(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}

	tmp_result = doca_dev_close(resources->doca_device);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to close DOCA device: %s", doca_error_get_descr(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}

	return result;
}

/**
 * @brief Destroy DPA thread
 *
 * @dpa_thread_obj [in]: DPA thread object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_thread_obj_destroy(struct dpa_thread_obj *dpa_thread_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, ret_err = DOCA_SUCCESS;

	doca_err = doca_dpa_thread_destroy(dpa_thread_obj->thread);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_destroy failed (%s)", doca_error_get_descr(doca_err));
		DOCA_ERROR_PROPAGATE(ret_err, doca_err);
	}

	return ret_err;
}

/**
 * @brief Initialize DPA thread
 *
 * @dpa_thread_obj [in/out]: DPA thread object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_thread_obj_init(struct dpa_thread_obj *dpa_thread_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS;

	doca_err = doca_dpa_thread_create(dpa_thread_obj->doca_dpa, &(dpa_thread_obj->thread));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_create failed (%s)", doca_error_get_descr(doca_err));
		return doca_err;
	}

	doca_err = doca_dpa_thread_set_func_arg(dpa_thread_obj->thread, dpa_thread_obj->func, dpa_thread_obj->arg);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_set_func_arg failed (%s)", doca_error_get_descr(doca_err));
		dpa_thread_obj_destroy(dpa_thread_obj);
		return doca_err;
	}

	doca_err = doca_dpa_thread_set_local_storage(dpa_thread_obj->thread, dpa_thread_obj->tls_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_set_local_storage failed (%s)", doca_error_get_descr(doca_err));
		dpa_thread_obj_destroy(dpa_thread_obj);
		return doca_err;
	}

	doca_err = doca_dpa_thread_start(dpa_thread_obj->thread);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_start failed (%s)", doca_error_get_descr(doca_err));
		dpa_thread_obj_destroy(dpa_thread_obj);
		return doca_err;
	}

	return doca_err;
}

/**
 * @brief Destroy DPA completion
 *
 * @dpa_completion_obj [in]: DPA completion object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_completion_obj_destroy(struct dpa_completion_obj *dpa_completion_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, ret_err = DOCA_SUCCESS;

	doca_err = doca_dpa_completion_destroy(dpa_completion_obj->dpa_comp);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_completion_destroy failed (%s)", doca_error_get_descr(doca_err));
		DOCA_ERROR_PROPAGATE(ret_err, doca_err);
	}

	return ret_err;
}

/**
 * @brief Initialize DPA completion
 *
 * @dpa_completion_obj [in/out]: DPA completion object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_completion_obj_init(struct dpa_completion_obj *dpa_completion_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS;

	doca_err = doca_dpa_completion_create(dpa_completion_obj->doca_dpa,
					      dpa_completion_obj->queue_size,
					      &(dpa_completion_obj->dpa_comp));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_completion_create failed (%s)", doca_error_get_descr(doca_err));
		return doca_err;
	}

	doca_err = doca_dpa_completion_set_thread(dpa_completion_obj->dpa_comp, dpa_completion_obj->thread);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_completion_set_thread failed (%s)", doca_error_get_descr(doca_err));
		dpa_completion_obj_destroy(dpa_completion_obj);
		return doca_err;
	}

	doca_err = doca_dpa_completion_start(dpa_completion_obj->dpa_comp);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_completion_start failed (%s)", doca_error_get_descr(doca_err));
		dpa_completion_obj_destroy(dpa_completion_obj);
		return doca_err;
	}

	doca_err = doca_dpa_completion_get_dpa_handle(dpa_completion_obj->dpa_comp, &(dpa_completion_obj->handle));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_completion_get_dpa_handle failed (%s)", doca_error_get_descr(doca_err));
		dpa_completion_obj_destroy(dpa_completion_obj);
		return doca_err;
	}

	return DOCA_SUCCESS;
}

/**
 * @brief Destroy DPA RDMA
 *
 * @dpa_rdma_obj [in]: DPA RDMA object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_rdma_obj_destroy(struct dpa_rdma_obj *dpa_rdma_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, ret_err = DOCA_SUCCESS;

	doca_err = doca_ctx_stop(dpa_rdma_obj->rdma_as_ctx);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_ctx_stop failed (%s)", doca_error_get_descr(doca_err));
		DOCA_ERROR_PROPAGATE(ret_err, doca_err);
	}

	doca_err = doca_rdma_destroy(dpa_rdma_obj->rdma);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_destroy failed (%s)", doca_error_get_descr(doca_err));
		DOCA_ERROR_PROPAGATE(ret_err, doca_err);
	}

	return ret_err;
}

/**
 * @brief Initialize DPA RDMA without starting it
 *
 * This function creates DPA RDMA object.
 * Please note that this function doesn't start the created DPA RDMA object, this need to be done
 * using dpa_rdma_obj_start() API
 *
 * @dpa_rdma_obj [in/out]: DPA RDMA object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_rdma_obj_init(struct dpa_rdma_obj *dpa_rdma_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS;

	doca_err = doca_rdma_create(dpa_rdma_obj->doca_device, &(dpa_rdma_obj->rdma));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_create failed (%s)", doca_error_get_descr(doca_err));
		return doca_err;
	}

	dpa_rdma_obj->rdma_as_ctx = doca_rdma_as_ctx(dpa_rdma_obj->rdma);

	doca_err = doca_rdma_set_permissions(dpa_rdma_obj->rdma, dpa_rdma_obj->permissions);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_set_permissions failed (%s)", doca_error_get_descr(doca_err));
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	doca_err = doca_rdma_set_grh_enabled(dpa_rdma_obj->rdma, true);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_set_grh_enabled failed (%s)", doca_error_get_descr(doca_err));
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	doca_err = doca_ctx_set_datapath_on_dpa(dpa_rdma_obj->rdma_as_ctx, dpa_rdma_obj->doca_dpa);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_ctx_set_datapath_on_dpa failed (%s)", doca_error_get_descr(doca_err));
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	doca_err = doca_ctx_set_user_data(dpa_rdma_obj->rdma_as_ctx, dpa_rdma_obj->user_data);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_ctx_set_user_data failed (%s)", doca_error_get_descr(doca_err));
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	return doca_err;
}

/**
 * @brief Start DPA RDMA
 *
 * @dpa_rdma_obj [in]: DPA RDMA object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t dpa_rdma_obj_start(struct dpa_rdma_obj *dpa_rdma_obj)
{
	/* start ctx */
	doca_error_t doca_err = doca_ctx_start(dpa_rdma_obj->rdma_as_ctx);
	if (doca_err != DOCA_SUCCESS) {
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	/* get dpa rdma handle */
	doca_err = doca_rdma_get_dpa_handle(dpa_rdma_obj->rdma, &(dpa_rdma_obj->dpa_rdma));
	if (doca_err != DOCA_SUCCESS) {
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	/* export connection details */
	doca_err = doca_rdma_export(dpa_rdma_obj->rdma,
				    &(dpa_rdma_obj->connection_details),
				    &(dpa_rdma_obj->conn_det_len));
	if (doca_err != DOCA_SUCCESS) {
		dpa_rdma_obj_destroy(dpa_rdma_obj);
		return doca_err;
	}

	return doca_err;
}

/**
 * @brief Create, attach, start and connect DPA RDMA objects
 *
 * This function creates DPA RDMA objects, attaches to DPA completion, start and connect them
 *
 * @resources [in]: DPA resources
 * @ping_dpa_completion_obj [in]: Ping DPA completion
 * @pong_dpa_completion_obj [in]: Pong DPA completion
 * @ping_dpa_rdma_obj [out]: Created ping DPA RDMA
 * @pong_dpa_rdma_obj [out]: Created pong DPA RDMA
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t rdma_objs_init(struct dpa_resources *resources,
				   struct dpa_completion_obj *ping_dpa_completion_obj,
				   struct dpa_completion_obj *pong_dpa_completion_obj,
				   struct dpa_rdma_obj *ping_dpa_rdma_obj,
				   struct dpa_rdma_obj *pong_dpa_rdma_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, tmp_doca_err = DOCA_SUCCESS;

	ping_dpa_rdma_obj->doca_device = resources->doca_device;
	ping_dpa_rdma_obj->doca_dpa = resources->doca_dpa;
	ping_dpa_rdma_obj->permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					 DOCA_ACCESS_FLAG_RDMA_READ;
	doca_err = dpa_rdma_obj_init(ping_dpa_rdma_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto exit;
	}

	pong_dpa_rdma_obj->doca_device = resources->doca_device;
	pong_dpa_rdma_obj->doca_dpa = resources->doca_dpa;
	pong_dpa_rdma_obj->permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					 DOCA_ACCESS_FLAG_RDMA_READ;
	doca_err = dpa_rdma_obj_init(pong_dpa_rdma_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_rdma_obj;
	}

	doca_err = doca_rdma_dpa_completion_attach(ping_dpa_rdma_obj->rdma, ping_dpa_completion_obj->dpa_comp);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_dpa_completion_attach failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	doca_err = doca_rdma_dpa_completion_attach(pong_dpa_rdma_obj->rdma, pong_dpa_completion_obj->dpa_comp);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_dpa_completion_attach failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	doca_err = dpa_rdma_obj_start(ping_dpa_rdma_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_start failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	doca_err = dpa_rdma_obj_start(pong_dpa_rdma_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_start failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	doca_err = doca_rdma_connect(ping_dpa_rdma_obj->rdma,
				     pong_dpa_rdma_obj->connection_details,
				     pong_dpa_rdma_obj->conn_det_len);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_connect failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	doca_err = doca_rdma_connect(pong_dpa_rdma_obj->rdma,
				     ping_dpa_rdma_obj->connection_details,
				     ping_dpa_rdma_obj->conn_det_len);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_rdma_connect failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_rdma_obj;
	}

	return doca_err;

destroy_pong_rdma_obj:
	tmp_doca_err = dpa_rdma_obj_destroy(pong_dpa_rdma_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_rdma_obj:
	tmp_doca_err = dpa_rdma_obj_destroy(ping_dpa_rdma_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

exit:
	return doca_err;
}

/**
 * @brief Destroy DPA RDMA objects
 *
 * @ping_dpa_rdma_obj [in]: Previously created ping DPA RDMA
 * @pong_dpa_rdma_obj [in]: Previously created pong DPA RDMA
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t rdma_objs_destroy(struct dpa_rdma_obj *ping_dpa_rdma_obj, struct dpa_rdma_obj *pong_dpa_rdma_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, tmp_doca_err = DOCA_SUCCESS;

	tmp_doca_err = dpa_rdma_obj_destroy(ping_dpa_rdma_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

	tmp_doca_err = dpa_rdma_obj_destroy(pong_dpa_rdma_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_rdma_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

	return doca_err;
}

/**
 * @brief Create DOCA sync event to be published by the DPA and subscribed by the CPU
 *
 * @doca_dpa [in]: DOCA DPA context
 * @doca_device [in]: DOCA device
 * @se [out]: Created DOCA sync event that is published by the DPA and subscribed by the CPU
 * @se_handle [out]: Created DOCA sync event handle
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_doca_dpa_completion_sync_event(struct doca_dpa *doca_dpa,
							  struct doca_dev *doca_device,
							  struct doca_sync_event **se,
							  doca_dpa_dev_sync_event_t *se_handle)
{
	doca_error_t result, tmp_result;

	result = doca_sync_event_create(se);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA sync event: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_sync_event_add_publisher_location_dpa(*se, doca_dpa);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set DPA as publisher for DOCA sync event: %s", doca_error_get_descr(result));
		goto destroy_comp_event;
	}

	result = doca_sync_event_add_subscriber_location_cpu(*se, doca_device);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set CPU as subscriber for DOCA sync event: %s", doca_error_get_descr(result));
		goto destroy_comp_event;
	}

	result = doca_sync_event_start(*se);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to start DOCA sync event: %s", doca_error_get_descr(result));
		goto destroy_comp_event;
	}

	result = doca_sync_event_get_dpa_handle(*se, doca_dpa, se_handle);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_sync_event_get_dpa_handle failed (%d)", result);
		goto destroy_comp_event;
	}

	return result;

destroy_comp_event:
	tmp_result = doca_sync_event_destroy(*se);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to destroy DOCA sync event: %s", doca_error_get_descr(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}
	return result;
}

/**
 * @brief Destroy DOCA Mmap
 *
 * @doca_mmap_obj [in]: DOCA Mmap object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t doca_mmap_obj_destroy(struct doca_mmap_obj *doca_mmap_obj)
{
	doca_error_t doca_err = DOCA_SUCCESS, ret_err = DOCA_SUCCESS;

	doca_err = doca_mmap_destroy(doca_mmap_obj->mmap);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_destroy failed (%s)", doca_error_get_descr(doca_err));
		DOCA_ERROR_PROPAGATE(ret_err, doca_err);
	}

	return ret_err;
}

/**
 * @brief Initialize DOCA Mmap
 *
 * @doca_mmap_obj [in/out]: DOCA Mmap object
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t doca_mmap_obj_init(struct doca_mmap_obj *doca_mmap_obj)
{
	doca_error_t doca_err = doca_mmap_create(&(doca_mmap_obj->mmap));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_create failed (%s)", doca_error_get_descr(doca_err));
		return doca_err;
	}

	doca_err = doca_mmap_set_permissions(doca_mmap_obj->mmap, doca_mmap_obj->permissions);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_set_permissions failed (%s)", doca_error_get_descr(doca_err));
		doca_mmap_obj_destroy(doca_mmap_obj);
		return doca_err;
	}

	doca_err = doca_mmap_set_dpa_memrange(doca_mmap_obj->mmap,
					      doca_mmap_obj->doca_dpa,
					      (uint64_t)doca_mmap_obj->memrange_addr,
					      doca_mmap_obj->memrange_len);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_set_dpa_memrange failed (%s)", doca_error_get_descr(doca_err));
		doca_mmap_obj_destroy(doca_mmap_obj);
		return doca_err;
	}

	doca_err = doca_mmap_start(doca_mmap_obj->mmap);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_start failed (%s)", doca_error_get_descr(doca_err));
		doca_mmap_obj_destroy(doca_mmap_obj);
		return doca_err;
	}

	doca_err = doca_mmap_dev_get_dpa_handle(doca_mmap_obj->mmap,
						doca_mmap_obj->doca_device,
						&(doca_mmap_obj->dpa_mmap_handle));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_dev_get_dpa_handle failed (%s)", doca_error_get_descr(doca_err));
		doca_mmap_obj_destroy(doca_mmap_obj);
		return doca_err;
	}

	doca_err = doca_mmap_export_rdma(doca_mmap_obj->mmap,
					 doca_mmap_obj->doca_device,
					 &(doca_mmap_obj->rdma_export),
					 &(doca_mmap_obj->export_len));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_export_rdma failed (%s)", doca_error_get_descr(doca_err));
		doca_mmap_obj_destroy(doca_mmap_obj);
		return doca_err;
	}

	return doca_err;
}

/**
 * @brief Application Logic
 *
 * @resources [in]: DPA resources that the application will use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t dpa_ping_pong(struct dpa_resources *resources)
{
	doca_error_t doca_err = DOCA_SUCCESS, tmp_doca_err = DOCA_SUCCESS;
	doca_dpa_dev_uintptr_t ping_thread_arg_dev_ptr = 0;
	doca_dpa_dev_uintptr_t ping_tls_dev_ptr = 0;
	doca_dpa_dev_uintptr_t ping_recv_arr_dev_ptr = 0;
	struct dpa_ping_pong_tls ping_tls = {0};
	doca_dpa_dev_uintptr_t pong_thread_arg_dev_ptr = 0;
	doca_dpa_dev_uintptr_t pong_tls_dev_ptr = 0;
	doca_dpa_dev_uintptr_t pong_recv_arr_dev_ptr = 0;
	struct dpa_ping_pong_tls pong_tls = {0};
	struct dpa_thread_obj ping_thread_obj = {0};
	struct dpa_thread_obj pong_thread_obj = {0};
	struct dpa_completion_obj ping_dpa_completion_obj = {0};
	struct dpa_completion_obj pong_dpa_completion_obj = {0};
	struct dpa_rdma_obj ping_rdma_obj = {0};
	struct dpa_rdma_obj pong_rdma_obj = {0};
	struct doca_sync_event *ping_comp_se = NULL;
	doca_dpa_dev_sync_event_t ping_comp_se_handle = 0;
	struct doca_sync_event *pong_comp_se = NULL;
	doca_dpa_dev_sync_event_t pong_comp_se_handle = 0;
	const uint64_t wait_sync_ev_threshold = 9;
	doca_dpa_dev_uintptr_t ping_receive_buf_dev_ptr = 0;
	struct doca_mmap_obj ping_receive_mmap_obj = {0};
	doca_dpa_dev_uintptr_t pong_receive_buf_dev_ptr = 0;
	struct doca_mmap_obj pong_receive_mmap_obj = {0};
	doca_dpa_dev_uintptr_t ping_send_buf_dev_ptr = 0;
	struct doca_mmap_obj ping_send_mmap_obj = {0};
	doca_dpa_dev_uintptr_t pong_send_buf_dev_ptr = 0;
	struct doca_mmap_obj pong_send_mmap_obj = {0};
	struct dpa_thread_arg ping_thread_arg = {0};
	struct dpa_thread_arg pong_thread_arg = {0};
	uint64_t retval = 0;

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(struct dpa_thread_arg), &ping_thread_arg_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto exit;
	}
	doca_err = doca_dpa_memset(resources->doca_dpa, ping_thread_arg_dev_ptr, 0, sizeof(struct dpa_thread_arg));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_thread_arg_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(struct dpa_ping_pong_tls), &ping_tls_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_thread_arg_dev_ptr;
	}
	doca_err = doca_dpa_mem_alloc(resources->doca_dpa,
				      sizeof(uint32_t) * EXPECTED_NUM_RECEIVES,
				      &ping_recv_arr_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_tls_dev_ptr;
	}
	ping_tls.received_values_arr_ptr = ping_recv_arr_dev_ptr;
	ping_tls.num_receives = 0;
	ping_tls.is_ping_thread = 1;
	doca_err =
		doca_dpa_h2d_memcpy(resources->doca_dpa, ping_tls_dev_ptr, &ping_tls, sizeof(struct dpa_ping_pong_tls));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_h2d_memcpy failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_recv_arr_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(struct dpa_thread_arg), &pong_thread_arg_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_recv_arr_dev_ptr;
	}
	doca_err = doca_dpa_memset(resources->doca_dpa, pong_thread_arg_dev_ptr, 0, sizeof(struct dpa_thread_arg));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_thread_arg_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(struct dpa_ping_pong_tls), &pong_tls_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_thread_arg_dev_ptr;
	}
	doca_err = doca_dpa_mem_alloc(resources->doca_dpa,
				      sizeof(uint32_t) * EXPECTED_NUM_RECEIVES,
				      &pong_recv_arr_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_tls_dev_ptr;
	}
	pong_tls.received_values_arr_ptr = pong_recv_arr_dev_ptr;
	pong_tls.num_receives = 0;
	pong_tls.is_ping_thread = 0;
	doca_err =
		doca_dpa_h2d_memcpy(resources->doca_dpa, pong_tls_dev_ptr, &pong_tls, sizeof(struct dpa_ping_pong_tls));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_h2d_memcpy failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_recv_arr_dev_ptr;
	}

	ping_thread_obj.doca_dpa = resources->doca_dpa;
	ping_thread_obj.func = &thread_kernel;
	ping_thread_obj.arg = ping_thread_arg_dev_ptr;
	ping_thread_obj.tls_dev_ptr = ping_tls_dev_ptr;
	doca_err = dpa_thread_obj_init(&ping_thread_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_thread_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_recv_arr_dev_ptr;
	}
	pong_thread_obj.doca_dpa = resources->doca_dpa;
	pong_thread_obj.func = &thread_kernel;
	pong_thread_obj.arg = pong_thread_arg_dev_ptr;
	pong_thread_obj.tls_dev_ptr = pong_tls_dev_ptr;
	doca_err = dpa_thread_obj_init(&pong_thread_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_thread_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_thread_obj;
	}

	ping_dpa_completion_obj.doca_dpa = resources->doca_dpa;
	ping_dpa_completion_obj.queue_size = 2;
	ping_dpa_completion_obj.thread = ping_thread_obj.thread;
	doca_err = dpa_completion_obj_init(&ping_dpa_completion_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_completion_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_thread_obj;
	}
	pong_dpa_completion_obj.doca_dpa = resources->doca_dpa;
	pong_dpa_completion_obj.queue_size = 2;
	pong_dpa_completion_obj.thread = pong_thread_obj.thread;
	doca_err = dpa_completion_obj_init(&pong_dpa_completion_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_completion_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_dpa_completion_obj;
	}

	doca_err = rdma_objs_init(resources,
				  &ping_dpa_completion_obj,
				  &pong_dpa_completion_obj,
				  &ping_rdma_obj,
				  &pong_rdma_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function rdma_objs_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_dpa_completion_obj;
	}

	doca_err = doca_dpa_thread_run(ping_thread_obj.thread);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_run failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_rdma_objs;
	}
	doca_err = doca_dpa_thread_run(pong_thread_obj.thread);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_thread_run failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_rdma_objs;
	}

	doca_err = create_doca_dpa_completion_sync_event(resources->doca_dpa,
							 resources->doca_device,
							 &ping_comp_se,
							 &ping_comp_se_handle);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function create_doca_dpa_completion_sync_event failed (%s)",
			     doca_error_get_descr(doca_err));
		goto destroy_rdma_objs;
	}
	doca_err = create_doca_dpa_completion_sync_event(resources->doca_dpa,
							 resources->doca_device,
							 &pong_comp_se,
							 &pong_comp_se_handle);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function create_doca_dpa_completion_sync_event failed (%s)",
			     doca_error_get_descr(doca_err));
		goto destoy_ping_completion_sync_event;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(uint64_t), &ping_receive_buf_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destoy_pong_completion_sync_event;
	}

	doca_err = doca_dpa_memset(resources->doca_dpa, ping_receive_buf_dev_ptr, 0, sizeof(uint64_t));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_receive_buf_dev_ptr;
	}

	ping_receive_mmap_obj.doca_dpa = resources->doca_dpa;
	ping_receive_mmap_obj.doca_device = resources->doca_device;
	ping_receive_mmap_obj.permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					    DOCA_ACCESS_FLAG_RDMA_READ;
	ping_receive_mmap_obj.memrange_addr = (void *)ping_receive_buf_dev_ptr;
	ping_receive_mmap_obj.memrange_len = sizeof(uint64_t);
	doca_err = doca_mmap_obj_init(&ping_receive_mmap_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_receive_buf_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(uint64_t), &pong_receive_buf_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_receive_mmap_obj;
	}

	doca_err = doca_dpa_memset(resources->doca_dpa, pong_receive_buf_dev_ptr, 0, sizeof(uint64_t));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_receive_buf_dev_ptr;
	}

	pong_receive_mmap_obj.doca_dpa = resources->doca_dpa;
	pong_receive_mmap_obj.doca_device = resources->doca_device;
	pong_receive_mmap_obj.permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					    DOCA_ACCESS_FLAG_RDMA_READ;
	pong_receive_mmap_obj.memrange_addr = (void *)pong_receive_buf_dev_ptr;
	pong_receive_mmap_obj.memrange_len = sizeof(uint64_t);
	doca_err = doca_mmap_obj_init(&pong_receive_mmap_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_receive_buf_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(uint64_t), &ping_send_buf_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_receive_mmap_obj;
	}

	doca_err = doca_dpa_memset(resources->doca_dpa, ping_send_buf_dev_ptr, 0, sizeof(uint64_t));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_send_buf_dev_ptr;
	}

	ping_send_mmap_obj.doca_dpa = resources->doca_dpa;
	ping_send_mmap_obj.doca_device = resources->doca_device;
	ping_send_mmap_obj.permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					 DOCA_ACCESS_FLAG_RDMA_READ;
	ping_send_mmap_obj.memrange_addr = (void *)ping_send_buf_dev_ptr;
	ping_send_mmap_obj.memrange_len = sizeof(uint64_t);
	doca_err = doca_mmap_obj_init(&ping_send_mmap_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_send_buf_dev_ptr;
	}

	doca_err = doca_dpa_mem_alloc(resources->doca_dpa, sizeof(uint64_t), &pong_send_buf_dev_ptr);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_alloc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_ping_send_mmap_obj;
	}

	doca_err = doca_dpa_memset(resources->doca_dpa, pong_send_buf_dev_ptr, 0, sizeof(uint64_t));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_memset failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_buf_dev_ptr;
	}

	pong_send_mmap_obj.doca_dpa = resources->doca_dpa;
	pong_send_mmap_obj.doca_device = resources->doca_device;
	pong_send_mmap_obj.permissions = DOCA_ACCESS_FLAG_LOCAL_READ_WRITE | DOCA_ACCESS_FLAG_RDMA_WRITE |
					 DOCA_ACCESS_FLAG_RDMA_READ;
	pong_send_mmap_obj.memrange_addr = (void *)pong_send_buf_dev_ptr;
	pong_send_mmap_obj.memrange_len = sizeof(uint64_t);
	doca_err = doca_mmap_obj_init(&pong_send_mmap_obj);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_init failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_buf_dev_ptr;
	}

	ping_thread_arg.dpa_comp_handle = ping_dpa_completion_obj.handle;
	ping_thread_arg.rdma_handle = (doca_dpa_dev_uintptr_t)ping_rdma_obj.dpa_rdma;
	ping_thread_arg.recv_addr = ping_receive_buf_dev_ptr;
	ping_thread_arg.recv_addr_mmap_handle = ping_receive_mmap_obj.dpa_mmap_handle;
	ping_thread_arg.recv_addr_length = ping_receive_mmap_obj.memrange_len;
	ping_thread_arg.send_addr = ping_send_buf_dev_ptr;
	ping_thread_arg.send_addr_mmap_handle = ping_send_mmap_obj.dpa_mmap_handle;
	ping_thread_arg.send_addr_length = ping_send_mmap_obj.memrange_len;
	ping_thread_arg.comp_sync_event_handle = ping_comp_se_handle;
	ping_thread_arg.comp_sync_event_val = wait_sync_ev_threshold + 1;
	doca_err = doca_dpa_h2d_memcpy(resources->doca_dpa,
				       ping_thread_arg_dev_ptr,
				       &ping_thread_arg,
				       sizeof(struct dpa_thread_arg));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_h2d_memcpy failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_mmap_obj;
	}

	pong_thread_arg.dpa_comp_handle = pong_dpa_completion_obj.handle;
	pong_thread_arg.rdma_handle = (doca_dpa_dev_uintptr_t)pong_rdma_obj.dpa_rdma;
	pong_thread_arg.recv_addr = pong_receive_buf_dev_ptr;
	pong_thread_arg.recv_addr_mmap_handle = pong_receive_mmap_obj.dpa_mmap_handle;
	pong_thread_arg.recv_addr_length = pong_receive_mmap_obj.memrange_len;
	pong_thread_arg.send_addr = pong_send_buf_dev_ptr;
	pong_thread_arg.send_addr_mmap_handle = pong_send_mmap_obj.dpa_mmap_handle;
	pong_thread_arg.send_addr_length = pong_send_mmap_obj.memrange_len;
	pong_thread_arg.comp_sync_event_handle = pong_comp_se_handle;
	pong_thread_arg.comp_sync_event_val = wait_sync_ev_threshold + 1;
	doca_err = doca_dpa_h2d_memcpy(resources->doca_dpa,
				       pong_thread_arg_dev_ptr,
				       &pong_thread_arg,
				       sizeof(struct dpa_thread_arg));
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_h2d_memcpy failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_mmap_obj;
	}

	doca_err = doca_dpa_rpc(resources->doca_dpa,
				&trigger_first_iteration_rpc,
				&retval,
				ping_thread_arg,
				pong_thread_arg);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_rpc failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_mmap_obj;
	}

	doca_err = doca_sync_event_wait_gt(ping_comp_se, wait_sync_ev_threshold, SYNC_EVENT_MASK_FFS);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_sync_event_wait_gt failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_mmap_obj;
	}

	doca_err = doca_sync_event_wait_gt(pong_comp_se, wait_sync_ev_threshold, SYNC_EVENT_MASK_FFS);
	if (doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_sync_event_wait_gt failed (%s)", doca_error_get_descr(doca_err));
		goto destroy_pong_send_mmap_obj;
	}

destroy_pong_send_mmap_obj:
	tmp_doca_err = doca_mmap_obj_destroy(&pong_send_mmap_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_send_buf_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, pong_send_buf_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_send_mmap_obj:
	tmp_doca_err = doca_mmap_obj_destroy(&ping_send_mmap_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_send_buf_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, ping_send_buf_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_receive_mmap_obj:
	tmp_doca_err = doca_mmap_obj_destroy(&pong_receive_mmap_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_receive_buf_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, pong_receive_buf_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_receive_mmap_obj:
	tmp_doca_err = doca_mmap_obj_destroy(&ping_receive_mmap_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_mmap_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_receive_buf_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, ping_receive_buf_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destoy_pong_completion_sync_event:
	tmp_doca_err = doca_sync_event_destroy(pong_comp_se);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_sync_event_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destoy_ping_completion_sync_event:
	tmp_doca_err = doca_sync_event_destroy(ping_comp_se);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_sync_event_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_rdma_objs:
	tmp_doca_err = rdma_objs_destroy(&ping_rdma_obj, &pong_rdma_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function rdma_objs_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_dpa_completion_obj:
	tmp_doca_err = dpa_completion_obj_destroy(&pong_dpa_completion_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_completion_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_dpa_completion_obj:
	tmp_doca_err = dpa_completion_obj_destroy(&ping_dpa_completion_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_completion_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_thread_obj:
	tmp_doca_err = dpa_thread_obj_destroy(&pong_thread_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_thread_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_thread_obj:
	tmp_doca_err = dpa_thread_obj_destroy(&ping_thread_obj);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function dpa_thread_obj_destroy failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_recv_arr_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, pong_recv_arr_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_tls_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, pong_tls_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_pong_thread_arg_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, pong_thread_arg_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_recv_arr_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, ping_recv_arr_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_tls_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, ping_tls_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

destroy_ping_thread_arg_dev_ptr:
	tmp_doca_err = doca_dpa_mem_free(resources->doca_dpa, ping_thread_arg_dev_ptr);
	if (tmp_doca_err != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Function doca_dpa_mem_free failed: %s", doca_error_get_descr(tmp_doca_err));
		DOCA_ERROR_PROPAGATE(doca_err, tmp_doca_err);
	}

exit:
	return doca_err;
}
